/*
     []-------------------------------------------------------------[]
     |	                     GEOMDIF.C                               |
     |				                                     |
     |     Funcoes, macros, constantes para o calculo de vetores     |
     |	        tangentes, curvatura e torcao de uma curva           |
     |						                     |
     |           por Lenimar N. Andrade, maio/96-abril/97            |
     []-------------------------------------------------------------[]
*/

#include <math.h>

#define PI	     3.14159265358979323846
#define PI_SOBRE_2   1.57079632679489661923
#define PI_SOBRE_4   0.78539816339744830961
#define DOIS_PI      6.28318530717958647692

/* Quadrado de x */
#define SQR(x)  ((x)*(x))

/* Distancia entre os pontos (x1, y1, z1) e (x2, y2, z2) */
#define DISTANCIA(x1,y1,z1,x2,y2,z2) sqrt(SQR((x1)-(x2)) + SQR((y1)-(y2)) \
							 + SQR((z1)-(z2)))
/* Determinante da matriz {{a, b}, {c, d}} */
#define Det2x2(a, b, c, d)  ((a)*(d) - (b)*(c))

/* Determinante da matriz {{a, b, c}, {d, e, f}, {g, h, i}} */
#define Det3x3(a,b,c,d,e,f,g,h,i)  ((a)*(e)*(i)+(c)*(d)*(h)+(b)*(f)*(g)\
				   -(c)*(e)*(g)-(a)*(f)*(h)-(b)*(d)*(i))
/* "Infinito" */
#define INF 3.4e38

/* f'(a) = derivada de f no ponto a */
#define DER(f,a,h)  ((f(a+h) - f(a-h))/(2.0*h))

/* f"(a) = derivada segunda de f no ponto a */
#define DER2(f,a,h) ((-2.0*f(a) + f(a-2.0*h) + f(a+2.0*h))/(4.0*h*h))

/* f'''(a) = derivada terceira de f no ponto a */
#define DER3(f,a,h) ((-f(a-3*h) + 3*f(a-h) - 3*f(a+h) + f(a+3*h))/(8.0*h*h*h))

/* Derivada parcial de f com relacao `a primeira variavel */
#define DIF1(f,a,b,h)  ((f(a+h,b) - f(a-h,b))/(2.0*h))

/* Derivada parcial de f com relacao `a segunda variavel */
#define DIF2(f,a,b,h)  ((f(a,b+h) - f(a,b-h))/(2.0*h))

/* Variaveis globais */
double h_erro = 1e-6,      /* Usado no calculo aproximado das derivadas    */
       z_erro = 1e-6;      /* Se |x| < z_erro, entao x sera' considerado 0 */
char mens_erro[100];       /* Mensagem de erro                             */
int houve_erro = 0;        /* Determina se houve erro na execucao          */

/* Produto vetorial (v1, v2, v3) = (a1, a2, a3) x (b1, b2, b3) */
#define PROD_VET(v1,v2,v3,a1,a2,a3,b1,b2,b3)   v1=(a2)*(b3)-(a3)*(b2); \
		       v2=(a3)*(b1)-(a1)*(b3); v3=(a1)*(b2)-(a2)*(b1)

/* Norma do vetor (v1, v2, v3) */
#define NORMA(v1, v2, v3)       (sqrt((v1)*(v1)+(v2)*(v2)+(v3)*(v3)))

typedef struct { double x, y, z; } vetor;       /* Vetor (x, y, z) no R3   */
typedef struct { double x, y, z; } ponto;       /* Ponto (x, y, z) no R3   */

/* ------------------------------------------------------------------------ */

double Curvatura(double (*f1)(double), double (*f2)(double),
		 double (*f3)(double), double t) {

/* Calcula a curvatura de f(t) = (f1(t), f2(t), f3(t)) */

   vetor v, d1, d2;
   double a, b;

   d1.x = DER(f1, t, h_erro);
   d1.y = DER(f2, t, h_erro);
   d1.z = DER(f3, t, h_erro);

   d2.x = DER2(f1, t, h_erro);
   d2.y = DER2(f2, t, h_erro);
   d2.z = DER2(f3, t, h_erro);

   PROD_VET(v.x, v.y, v.z, d1.x, d1.y, d1.z, d2.x, d2.y, d2.z);
   a = NORMA(v.x, v.y, v.z);
   b = NORMA(d1.x, d1.y, d1.z);
   b = b*b*b;

   if (fabs(b) < z_erro) return INF; else return a/b;
}

/* ----------------------------------------------------------------------- */

void VetorTangenteACurva(vetor *v, double (*f1)(double), double (*f2)(double),
			double (*f3)(double), double t0) {

/* Calcula o vetor tangente `a curva (f1(t), f2(t), f3(t)) no ponto t = t0 */

  v->x = DER(f1, t0, h_erro);
  v->y = DER(f2, t0, h_erro);
  v->z = DER(f3, t0, h_erro);
}

/* ----------------------------------------------------------------------- */

double Dist(ponto P, double (*f1)(double), double (*f2)(double),
       double (*f3)(double), double a, double b, double incr, double *tmin) {

/* Calcula a distancia do ponto P `a curva f(t) = (f1(t), f2(t), f3(t)),
   a <= t <= b, usando incremento incr em t */

   double t, d, dmin = INF;

   t = a;
   do {
     d = DISTANCIA(P.x, P.y, P.z, f1(t), f2(t), f3(t));
     if (d < dmin) { dmin = d; *tmin = t; }
     t += incr;
   } while (t <= b);

   return dmin;
}

/* ----------------------------------------------------------------------- */

double Torcao(double (*f1)(double), double (*f2)(double),
		 double (*f3)(double), double t) {

/* Calcula a torcao de f(t) = (f1(t), f2(t), f3(t)) */

   vetor v, d1, d2, d3;
   double a, b;

   d1.x = DER(f1, t, h_erro);
   d1.y = DER(f2, t, h_erro);
   d1.z = DER(f3, t, h_erro);

   d2.x = DER2(f1, t, h_erro);
   d2.y = DER2(f2, t, h_erro);
   d2.z = DER2(f3, t, h_erro);

   d3.x = DER3(f1, t, h_erro);
   d3.y = DER3(f2, t, h_erro);
   d3.z = DER3(f3, t, h_erro);

   PROD_VET(v.x, v.y, v.z, d1.x, d1.y, d1.z, d2.x, d2.y, d2.z);

   a = v.x*d3.x + v.y*d3.y + v.z*d3.z;

   b = NORMA(v.x, v.y, v.z);
   b = b*b;

   if (fabs(b) < z_erro) return INF; else return a/b;
}

/* ----------------------------------------------------------------------- */

void CentroDeCurvatura(double (*f1) (double), double (*f2) (double),
    double (*f3) (double), double a, double *k1, double *k2, double *k3) {

/* Calcula o centro do circulo osculador de f(t) = (f1(t), f2(t), f3(t))
   quando t = a */

  *k1 = (-(f1(a) - f1(a - 2*h_erro))/
      (2*h_erro*sqrt(SQR(f1(a) - f1(a - 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(f2(a) - f2(a - 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(f3(a) - f3(a - 2*h_erro))/(4*h_erro*h_erro))) +
     (-f1(a) + f1(a + 2*h_erro))/
      (2*h_erro*sqrt(SQR(-f1(a) + f1(a + 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(-f2(a) + f2(a + 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(-f3(a) + f3(a + 2*h_erro))/(4*h_erro*h_erro))))/
   (2*h_erro*sqrt(SQR(-f1(a - h_erro) + f1(a + h_erro))/(4*h_erro*h_erro) +
	SQR(-f2(a - h_erro) + f2(a + h_erro))/(4*h_erro*h_erro) +
	SQR(-f3(a - h_erro) + f3(a + h_erro))/(4*h_erro*h_erro)));

  *k2 = (-(f2(a) - f2(a - 2*h_erro))/
      (2*h_erro*sqrt(SQR(f1(a) - f1(a - 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(f2(a) - f2(a - 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(f3(a) - f3(a - 2*h_erro))/(4*h_erro*h_erro))) +
     (-f2(a) + f2(a + 2*h_erro))/
      (2*h_erro*sqrt(SQR(-f1(a) + f1(a + 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(-f2(a) + f2(a + 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(-f3(a) + f3(a + 2*h_erro))/(4*h_erro*h_erro))))/
   (2*h_erro*sqrt(SQR(-f1(a - h_erro) + f1(a + h_erro))/(4*h_erro*h_erro) +
	SQR(-f2(a - h_erro) + f2(a + h_erro))/(4*h_erro*h_erro) +
	SQR(-f3(a - h_erro) + f3(a + h_erro))/(4*h_erro*h_erro)));

  *k3 = (-(f3(a) - f3(a - 2*h_erro))/
      (2*h_erro*sqrt(SQR(f1(a) - f1(a - 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(f2(a) - f2(a - 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(f3(a) - f3(a - 2*h_erro))/(4*h_erro*h_erro))) +
     (-f3(a) + f3(a + 2*h_erro))/
      (2*h_erro*sqrt(SQR(-f1(a) + f1(a + 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(-f2(a) + f2(a + 2*h_erro))/(4*h_erro*h_erro) +
	   SQR(-f3(a) + f3(a + 2*h_erro))/(4*h_erro*h_erro))))/
   (2*h_erro*sqrt(SQR(-f1(a - h_erro) + f1(a + h_erro))/(4*h_erro*h_erro) +
	SQR(-f2(a - h_erro) + f2(a + h_erro))/(4*h_erro*h_erro) +
	SQR(-f3(a - h_erro) + f3(a + h_erro))/(4*h_erro*h_erro)));
}

/* ----------------------------------------------------------------------- */

/*** FIM DE GEOMDIF.C ***/
