/*** TST_CURV.C ***/

/* Calcula o centro do circulo osculador exato e o centro do circulo
   osculador aproximado de uma curva f(t) = (f1(t), f2(t), f3(t)) */

#include <math.h>
#include <stdio.h>

/* Definicao da curva */
/* ------------------ */
double f1(double t) { return sin(t)*t; }
double f2(double t) { return cos(t)/(3 + 4*t*t*t*t); }
double f3(double t) { return (t*t*t - 3*t*t + 1)/(5 + 2*cos(t)); }

/* Quadrado de x */
#define SQR(x)  ((x)*(x))

/* "Infinito" */
#define INF 3.4e38

/* Distancia entre os pontos (x1, y1, z1) e (x2, y2, z2) */
#define DISTANCIA(x1,y1,z1,x2,y2,z2) sqrt(SQR((x1)-(x2)) + SQR((y1)-(y2)) \
							 + SQR((z1)-(z2)))
/* f'(a) = derivada de f no ponto a */
#define DER(f,a,h)  ((f(a+h) - f(a-h))/(2.0*h))

/* f"(a) = derivada segunda de f no ponto a */
#define DER2(f,a,h) ((-2.0*f(a) + f(a-2.0*h) + f(a+2.0*h))/(4.0*h*h))

double h_erro = 1e-6,      /* Usado no calculo aproximado das derivadas    */
       z_erro = 1e-6;      /* Se |x| < z_erro, entao x sera' considerado 0 */

/* Produto vetorial (v1, v2, v3) = (a1, a2, a3) x (b1, b2, b3) */
#define PROD_VET(v1,v2,v3,a1,a2,a3,b1,b2,b3)   v1=(a2)*(b3)-(a3)*(b2); \
		       v2=(a3)*(b1)-(a1)*(b3); v3=(a1)*(b2)-(a2)*(b1)

/* Norma do vetor (v1, v2, v3) */
#define NORMA(v1, v2, v3)       (sqrt((v1)*(v1)+(v2)*(v2)+(v3)*(v3)))

typedef struct { double x, y, z; } vetor;       /* Vetor (x, y, z) no R3   */

vetor N, u, v;

/* ----------------------------------------------------------------------- */

double Curvatura(double (*f1)(double), double (*f2)(double),
		 double (*f3)(double), double t) {

/* Calcula a curvatura de f(t) = (f1(t), f2(t), f3(t)) */

   vetor v, d1, d2;
   double a, b;

   d1.x = DER(f1, t, h_erro);
   d1.y = DER(f2, t, h_erro);
   d1.z = DER(f3, t, h_erro);

   d2.x = DER2(f1, t, h_erro);
   d2.y = DER2(f2, t, h_erro);
   d2.z = DER2(f3, t, h_erro);

   PROD_VET(v.x, v.y, v.z, d1.x, d1.y, d1.z, d2.x, d2.y, d2.z);
   a = NORMA(v.x, v.y, v.z);
   b = NORMA(d1.x, d1.y, d1.z);
   b = b*b*b;

   if (fabs(b) < z_erro) return INF; else return a/b;
}

/* ----------------------------------------------------------------------- */

void Evoluta(double (*f1) (double), double (*f2) (double),
    double (*f3) (double), double a, double *C1, double *C2, double *C3) {

/* Calcula o centro do circulo osculador de f(t) = (f1(t), f2(t), f3(t))
   quando t = a. E' utilizada a formula:

   Evoluta(a) = f(a) + |f'(a)|^3/|f'(a) x f"(a)| * N(a)/|N(a)|

   onde N(a) = f"(a)/|f'(a)| - ((f'(a) . f"(a))/|f'(a)|^3) f'(a)   */

   vetor Df   /*= f'(a)*/,   D2f   /*= f"(a)*/,
	 prod_vet  /*= f'(a)xf"(a)*/;
   double R, prod_interno, norma_N, norma_Df, norma_DfxD2f;

   Df.x = DER(f1, a, h_erro),   Df.y = DER(f2, a, h_erro),
   Df.z = DER(f3, a, h_erro),
   norma_Df = NORMA(Df.x, Df.y, Df.z);

   D2f.x = DER2(f1, a, h_erro),   D2f.y = DER2(f2, a, h_erro),
   D2f.z = DER2(f3, a, h_erro),
   prod_interno = Df.x*D2f.x + Df.y*D2f.y + Df.z*D2f.z;
   PROD_VET(prod_vet.x, prod_vet.y, prod_vet.z,
				 Df.x, Df.y, Df.z, D2f.x, D2f.y, D2f.z);
   norma_DfxD2f = NORMA(prod_vet.x, prod_vet.y, prod_vet.z);

   R = norma_Df*norma_Df*norma_Df/norma_DfxD2f;
   N.x = D2f.x/norma_Df - Df.x*prod_interno/(norma_Df*norma_Df*norma_Df),
   N.y = D2f.y/norma_Df - Df.y*prod_interno/(norma_Df*norma_Df*norma_Df),
   N.z = D2f.z/norma_Df - Df.z*prod_interno/(norma_Df*norma_Df*norma_Df);
   norma_N = NORMA(N.x, N.y, N.z);

   *C1 = f1(a) + R*N.x/norma_N, *C2 = f2(a) + R*N.y/norma_N,
   *C3 = f3(a) + R*N.z/norma_N;
}

/* ----------------------------------------------------------------------- */

void CentroDeCurvaturaAprox(double (*f1) (double), double (*f2) (double),
				 double (*f3) (double), double a, double L,
				 double *c1, double *c2, double *c3) {

/* Calcula o centro do circulo osculador aproximado, resolvendo o sistema
   linear 3 x 3:
		  {  X . u = P . u
		  {  X . v = Q . v
		  {  X . (u x v) = Q . (u x v)

   onde X = (c1, c2, c3),
	P = (f1(a), f2(a), f3(a)),
	u = (f1'(a), f2'(a), f3'(a)),
	Q = (f1(a + L), f2(a + L), f3(a + L)),
	v = (f1'(a + L), f2'(a + L), f3'(a + L)) */

  double u1, u2, u3, v1, v2, v3, p1, p2, p3, q1, q2, q3;

  v1 = DER(f1,a+L,h_erro), v2 = DER(f2,a+L,h_erro), v3 = DER(f3,a+L,h_erro);
  u1 = DER(f1,a,h_erro), u2 = DER(f2,a,h_erro), u3 = DER(f3,a,h_erro);
  q1 = f1(a+L), q2 = f2(a+L), q3 = f3(a+L);
  p1 = f1(a), p2 = f2(a), p3 = f3(a);
  u.x = u1, u.y = u2, u.z = u3;
  v.x = v1, v.y = v2, v.z = v3;

*c1=(p1*u1*(u1*(SQR(v2)+SQR(v3))-v1*(u2*v2+u3*v3))+p2*u2*(u1*(SQR(v2)+SQR(v3))-
v1*(u2*v2+u3*v3))+p3*u3*(u1*(SQR(v2)+SQR(v3))-v1*(u2*v2+u3*v3))-q1*(u1*v1*(u2*
v2+u3*v3)-SQR(u2)*(SQR(v1)+SQR(v3))+u3*(2*u2*v2*v3-u3*(SQR(v1)+SQR(v2))))+(v1*
(u2*v2+u3*v3)-u1*(SQR(v2)+SQR(v3)))*(q2*u2+q3*u3))/(SQR(u1)*(SQR(v2)+SQR(v3))-
2*u1*v1*(u2*v2+u3*v3)+SQR(u2)*(SQR(v1)+SQR(v3))-2*u2*u3*v2*v3+SQR(u3)*(SQR(v1)
+SQR(v2)));

*c2=-(p1*u1*(u1*v1*v2-u2*(SQR(v1)+SQR(v3))+u3*v2*v3)+p2*u2*(u1*v1*v2-u2*(SQR(v1
)+SQR(v3))+u3*v2*v3)+p3*u3*(u1*v1*v2-u2*(SQR(v1)+SQR(v3))+u3*v2*v3)-q1*u1*(u1*
v1*v2-u2*(SQR(v1)+SQR(v3))+u3*v2*v3)-q2*(SQR(u1)*(SQR(v2)+SQR(v3))-u1*v1*(u2*v2
+2*u3*v3)-u2*u3*v2*v3+SQR(u3)*(SQR(v1)+SQR(v2)))-q3*u3*(u1*v1*v2-u2*(SQR(v1)+
SQR(v3))+u3*v2*v3))/(SQR(u1)*(SQR(v2)+SQR(v3))-2*u1*v1*(u2*v2+u3*v3)+SQR(u2)*(
SQR(v1)+SQR(v3))-2*u2*u3*v2*v3+SQR(u3)*(SQR(v1)+SQR(v2)));

*c3=-(p1*u1*(u1*v1*v3+u2*v2*v3-u3*(SQR(v1)+SQR(v2)))+p2*u2*(u1*v1*v3+u2*v2*v3-
u3*(SQR(v1)+SQR(v2)))+p3*u3*(u1*v1*v3+u2*v2*v3-u3*(SQR(v1)+SQR(v2)))-q1*u1*(u1*
v1*v3+u2*v2*v3-u3*(SQR(v1)+SQR(v2)))-q2*u2*(u1*v1*v3+u2*v2*v3-u3*(SQR(v1)+SQR(
v2)))-q3*(SQR(u1)*(SQR(v2)+SQR(v3))-u1*v1*(2*u2*v2+u3*v3)+u2*(u2*(SQR(v1)+SQR(
v3))-u3*v2*v3)))/(SQR(u1)*(SQR(v2)+SQR(v3))-2*u1*v1*(u2*v2+u3*v3)+SQR(u2)*(SQR
(v1)+SQR(v3))-2*u2*u3*v2*v3+SQR(u3)*(SQR(v1)+SQR(v2)));

}

/* ----------------------------------------------------------------------- */

void main(void) {

 double a, a_min = -1, a_max = 1, a_incr = 0.5, L, C1, C2, C3, x1, x2, x3,
	v_interno_n;

 a = a_min;
 do {
   Evoluta(f1, f2, f3, a, &C1, &C2, &C3);
   printf(
   "\n\n=======================================================================");
   printf("\nCentro do circulo osculador [Exato] = (%f, %f, %f)", C1, C2, C3);
   printf("\nt = %f   Raio de curvatura [Exato] = %f", a, 1.0/Curvatura(f1, f2, f3, a));
   printf(
   "\n-----------------------------------------------------------------------");
   printf("\n    L             Centro [Aprox]             Raio         v.n");
   printf(
   "\n-----------------------------------------------------------------------");
   L = 1;
   do {
     printf("\n%f ", L);
     CentroDeCurvaturaAprox(f1, f2, f3, a, L, &x1, &x2, &x3);

     v_interno_n = N.x*v.x + N.y*v.y + N.z*v.z;

     printf(" (%f, %f, %f) ", x1, x2, x3);
     printf(" %f ", DISTANCIA(x1, x2, x3, f1(a), f2(a), f3(a)));
     printf("  %8.8f", v_interno_n);
     L /= 10.0;
   } while (L > 0.000001);
   printf(
   "\n=======================================================================\n");

  a += a_incr;
  } while (a <= a_max);
}

/*** FIM DE TST_CURV.C ***/
